
--
--  Gossipmonger
--    by Tuhljin
--

local THIS_VERSION = GetAddOnMetadata("Gossipmonger", "Version");
local THIS_TITLE = GetAddOnMetadata("Gossipmonger", "Title")

local L = GOSSIPMONGER_STRINGS

local CYAN = "|cff00ffff"
local YELLOW = "|cffffff00"
local YELLOW_LT = "|cffffff99"
local R, G, B = 0.494, 0.729, 1
--local SCTColors

local minimumGossipHeight, maximumGossipHeight = 150, 414
local incGossipHeight = 119

local ChatIcon = "|TInterface\\GossipFrame\\GossipGossipIcon:0:0:2:-2|t"

Gossipmonger = {};

Gossipmonger.DefaultSettings = {
  Version = THIS_VERSION,
  skipIsDefault = true,
  --informOnSkip_SCT = false,
  informOnSkip_log = false,
  alwaysInformWithName = false,
  showGossipMirrorOnSkip = true,
  delayShowingMirror = true,
  showGossipMirrorOnQuestSelect = true,
  showGossipMirrorOnQuestJump = true,
  greetingModifierKey = 2,
  Filter_Standard = true,
};

local CurrentNPC, CurrentNPC_gossip
local SafeMirrorOptionsClick
local GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip  -- For the GossipFrame_OnEvent hook
local IgnoreQuestFinished, FollowupFunction_filtername
local KeySelectDropDown
local typeReplacement

local function chatprint(msg)
  DEFAULT_CHAT_FRAME:AddMessage(msg, R, G, B);
end

local TQD, preTQD

local function prepGossipOptionDisplay(id, baseButton, selected, QuestGreeting)
  local name = "Gossipmonger_OptionDisplay"..id
  local titleButton = _G[name]
  if (not titleButton) then
    titleButton = CreateFrame("Button", name, Gossipmonger_MirrorFrame_GreetingScrollFrameScrollChild, "Gossipmonger_OptionDisplay")
    if (id == 1) then
      titleButton:SetPoint("TOPLEFT", Gossipmonger_MirrorFrame_GreetingText, "BOTTOMLEFT", -10, -20)
    else
      local prevButton = _G["Gossipmonger_OptionDisplay"..(id - 1)]
      titleButton:SetPoint("TOPLEFT", prevButton, "BOTTOMLEFT", 0, -3)
    end
  end

  titleButton.baseButton = baseButton
  local baseIcon = QuestGreeting and _G[baseButton.."QuestIcon"] or _G[baseButton.."GossipIcon"]
  titleButton.QuestGreeting = QuestGreeting
  baseButton = _G[baseButton]

  if (baseButton:IsShown()) then
    titleButton:Show()

    if (TRIVIAL_QUEST_DISPLAY ~= preTQD) then
      TQD = TRIVIAL_QUEST_DISPLAY:gsub("%(","%%(")
      TQD = TQD:gsub("%)","%%)")
      TQD = ("^"..TQD.."$"):gsub("%%s", "(.+)")
      preTQD = TRIVIAL_QUEST_DISPLAY
    end
    local NQD = ("^"..NORMAL_QUEST_DISPLAY.."$"):gsub("%%s", "(.+)")
    local text = baseButton:GetText()
    local _, baseTitle, taggedTitle
    _, _, baseTitle = text:find(TQD)
    _, _, taggedTitle = text:find(NQD)
    if (baseTitle) then  -- Low level quest
      titleButton.isLowLevel = true
    elseif (taggedTitle) then  -- Normal quest
      titleButton.isLowLevel = nil
      baseTitle = taggedTitle
    else
      baseTitle, taggedTitle = text, text
    end
    titleButton.title = baseTitle

    local title = selected and YELLOW .. taggedTitle .. "   " .. YELLOW_LT .. L.SELECTEDOPTION or
                  YELLOW .. taggedTitle
    titleButton:SetText(title);
    GossipResize(titleButton);   -- Calls a Blizzard function that currently simply does this:  titleButton:SetHeight( titleButton:GetTextHeight() + 2);
    titleButton:SetID( baseButton:GetID() )
    titleButton.type = baseButton.type  -- Set to "Available", "Active", or "Gossip"
    local icon = _G[name.."GossipIcon"]
    icon:SetTexture( baseIcon:GetTexture() )
    icon:SetVertexColor( baseIcon:GetVertexColor() )
    titleButton.selected = selected
    return titleButton:GetHeight(), true
  else
    titleButton.selected = nil
    titleButton:Hide()
  end
end

local function showGossipMirror(delay, name, text, selected, gossipType, noUpdate, watchQuest, numQuests, QuestGreeting)
  SafeMirrorOptionsClick = nil
  Gossipmonger_MirrorFrame_GreetingText:SetText(text);
  if (not noUpdate) then  -- noUpdate should be true when ShowGossipFrame already handled this
    GossipFrameUpdate() -- Update standard buttons; we'll grab the data from them rather than reinventing the
                        -- wheel. (It's hopefully more forward compatible this way, as well.)
  end
  -- Show used buttons:
  local optionsHeight, mirrorbtnIndex, h = -3, 1
  local buttonName = QuestGreeting and "QuestTitleButton" or "GossipTitleButton"
  local anySelected
  for i = 1, (QuestGreeting and numQuests or GossipFrame.buttonIndex - 1) do
  -- When using GossipFrame.buttonIndex, the actual number of buttons displayed is one less than given.
    local thisSelected = selected == mirrorbtnIndex
	if (thisSelected) then  anySelected = true;  end
    h = prepGossipOptionDisplay(mirrorbtnIndex, buttonName..i, thisSelected, QuestGreeting)
    if (h) then
      optionsHeight = optionsHeight + h + 3
      mirrorbtnIndex = mirrorbtnIndex + 1
    end
  end

  if (not anySelected) then -- Failed to find the selected option. Have QuestHanger do its search. (Failsafe for when selected variable doesn't match anything. Not sure exact circumstance it's needed, but it was for a Wanted Board object at Refuge Pointe that didn't have a proper NPC associated with it.)
    Gossipmonger_MirrorFrame.selected = nil
  else
    Gossipmonger_MirrorFrame.selected = selected
  end
  Gossipmonger_MirrorFrame.numOptions = mirrorbtnIndex - 1
  Gossipmonger_MirrorFrame.numQuests = numQuests or GetNumGossipAvailableQuests() + GetNumGossipActiveQuests()

  -- Hide the rest:
  local titleButton = _G["Gossipmonger_OptionDisplay"..mirrorbtnIndex]
  while (titleButton) do
    titleButton:Hide()
    mirrorbtnIndex = mirrorbtnIndex + 1
    titleButton = _G["Gossipmonger_OptionDisplay"..mirrorbtnIndex]
  end

  h = Gossipmonger_MirrorFrame_GreetingText:GetHeight() or 0
  local h2 = h + optionsHeight + incGossipHeight
  if (h2 < minimumGossipHeight) then
    Gossipmonger_MirrorFrame:SetHeight(minimumGossipHeight)
    Gossipmonger_MirrorFrame_GreetingScrollFrameScrollBar:Hide()
    Gossipmonger_MirrorFrame_GreetingScrollFrame:EnableMouseWheel(false)
  elseif (h <= Gossipmonger_MirrorFrame_GreetingScrollFrame:GetHeight() and h2 <= maximumGossipHeight) then
    Gossipmonger_MirrorFrame:SetHeight(h2)
    Gossipmonger_MirrorFrame_GreetingScrollFrameScrollBar:Hide()
    Gossipmonger_MirrorFrame_GreetingScrollFrame:EnableMouseWheel(false)
  else
    Gossipmonger_MirrorFrame:SetHeight(maximumGossipHeight)
    Gossipmonger_MirrorFrame_GreetingScrollFrameScrollBar:Show()
    Gossipmonger_MirrorFrame_GreetingScrollFrame:EnableMouseWheel()
  end

  CurrentNPC, CurrentNPC_gossip = name, gossipType

  Gossipmonger_MirrorFrame_NpcNameText:SetText( CurrentNPC or CYAN..THIS_TITLE )
  if ( UnitExists("npc") ) then
    Gossipmonger_MirrorFrame_NpcNameText:SetTextColor(GameTooltip_UnitColor("npc"))
    SetPortraitTexture(Gossipmonger_MirrorFrame_Portrait, "npc");
    Gossipmonger_MirrorFrame.NPCName = CurrentNPC
  else
    Gossipmonger_MirrorFrame_Portrait:SetTexture("Interface\\QuestFrame\\UI-QuestLog-BookIcon");
    Gossipmonger_MirrorFrame.NPCName = CurrentNPC --nil
  end

  Gossipmonger_MirrorFrame_GreetingScrollFrame:SetVerticalScroll(0);
  if (not delay) then
    ShowUIPanel(Gossipmonger_MirrorFrame)
  end

  if (watchQuest) then
    if (watchQuest == true or Gossipmonger_Settings.showGossipMirrorOnQuestSelect) then
	-- Old comment preserved (for now) in case I need it later, but I'm not sure it's accurate or reflects what I want any more:
    -- watchQuest is set to 1 if the mirror should only automatically pops up if the settings tell it to do so,
    -- or true if it should always pop up (as is the case when a filter selects it). Either way, if the player
    -- clicked the vendor option, for example, instead of a quest, the mirror shouldn't pop up since it wasn't
    -- due to a filter selecting it.
      Gossipmonger_MirrorFrame:RegisterEvent("QUEST_DETAIL")
      Gossipmonger_MirrorFrame:RegisterEvent("QUEST_PROGRESS")
      Gossipmonger_MirrorFrame:RegisterEvent("QUEST_COMPLETE")
	  --Gossipmonger_MirrorFrame:RegisterEvent("QUEST_ACCEPTED") --Don't register it here. The way we do it, has to be after QUEST_DETAIL triggers or things break.
    end
	--[[
  else
    Gossipmonger_MirrorFrame:RegisterEvent("BANKFRAME_OPENED")
    Gossipmonger_MirrorFrame:RegisterEvent("BATTLEFIELDS_SHOW")
    Gossipmonger_MirrorFrame:RegisterEvent("MERCHANT_SHOW")
    Gossipmonger_MirrorFrame:RegisterEvent("TAXIMAP_OPENED")
    Gossipmonger_MirrorFrame:RegisterEvent("TRAINER_SHOW")
    Gossipmonger_MirrorFrame:RegisterEvent("CONFIRM_BINDER")
	--]]
  end
  Gossipmonger_MirrorFrame:RegisterEvent("BANKFRAME_OPENED")
  Gossipmonger_MirrorFrame:RegisterEvent("BATTLEFIELDS_SHOW")
  Gossipmonger_MirrorFrame:RegisterEvent("MERCHANT_SHOW")
  Gossipmonger_MirrorFrame:RegisterEvent("TAXIMAP_OPENED")
  Gossipmonger_MirrorFrame:RegisterEvent("TRAINER_SHOW")
  Gossipmonger_MirrorFrame:RegisterEvent("CONFIRM_BINDER")
  Gossipmonger_MirrorFrame:RegisterEvent("AUCTION_HOUSE_SHOW")
end

local function BreakMirror()
-- We're not just hiding the mirror. We're "breaking" it away from the current NPC. (There's a distinction because
-- there may be times we want to have the option of showing it again w/o the disassociation that goes on here.)
  CurrentNPC, CurrentNPC_gossip, SafeMirrorOptionsClick = nil, nil, nil
  IgnoreQuestFinished = nil
  Gossipmonger_MirrorFrame.NPCName = nil
  if (Gossipmonger_MirrorFrame:IsShown()) then
    HideUIPanel(Gossipmonger_MirrorFrame)
  end
  Gossipmonger_MirrorFrame:UnregisterAllEvents();
end

local function ShowMirror()
  if (not Gossipmonger_MirrorFrame:IsShown()) then
    ShowUIPanel(Gossipmonger_MirrorFrame)
  end
  SafeMirrorOptionsClick = true

  Gossipmonger_MirrorFrame:UnregisterAllEvents();
  Gossipmonger_MirrorFrame:RegisterEvent("BANKFRAME_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("BATTLEFIELDS_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("MERCHANT_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("TAXIMAP_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("TRAINER_CLOSED")
  Gossipmonger_MirrorFrame:RegisterEvent("QUEST_FINISHED")
  Gossipmonger_MirrorFrame:RegisterEvent("AUCTION_HOUSE_CLOSED")
end

local function tooltip_StandardFilter(self, tooltip)
  tooltip:AddLine(L.FILTEROPT_STANDARD_TIP, nil, nil, nil, 1);
  tooltip:AddLine(L.FILTEROPT_STANDARD_TIP2, nil, nil, nil, 1);
  tooltip:AddLine(L.FILTEROPT_STANDARD_TIP3, nil, nil, nil, 1);
  tooltip:AddTexture("Interface\\GossipFrame\\GossipGossipIcon");
  tooltip:AddLine(L.FILTEROPT_STANDARD_TIP4, nil, nil, nil, 1);
end

local function UpdateLabel()
  local KEYTIP
  if (Gossipmonger_Settings.skipIsDefault) then
    Gossipmonger_GreetingModDropdown:SetLabel(L.MODIFIERKEY_LABEL_DONOT)
	KEYTIP = L.MODIFIERKEY_TOOLTIP_KEY_DONOT
  else
    Gossipmonger_GreetingModDropdown:SetLabel(L.MODIFIERKEY_LABEL_DO)
    KEYTIP = L.MODIFIERKEY_TOOLTIP_KEY_DO
  end
  for k,v in ipairs(KeySelectDropDown) do
    if (v.key) then
      v.tooltipText = KEYTIP:format(v.key)
	end
  end
end

local function UpdateStandardFilters()
  if (Gossipmonger_Settings.Filter_Standard) then
    for i,func in ipairs(Gossipmonger.StandardFilters) do
      Gossipmonger.EnableFilter(func)
    end
  else
    for i,func in ipairs(Gossipmonger.StandardFilters) do
      Gossipmonger.DisableFilter(func)
    end
  end
end

local function Init()
  Gossipmonger_MirrorFrame:SetScript("OnEvent", Gossipmonger.OnEvent);

  --local KEY, KEYTIP = L.KEY, L.MODIFIERKEY_TOOLTIP_KEY
  local KEY = L.KEY
  KeySelectDropDown = {
    {
      text = KEY:format(L.ALT),
      tooltipTitle = KEY:format(L.ALT),
      --tooltipText = KEYTIP:format(L.ALT),
	  key = L.ALT,
      value = 0
    },
    {
      text = KEY:format(L.CTRL),
      tooltipTitle = KEY:format(L.CTRL),
      --tooltipText = KEYTIP:format(L.CTRL),
	  key = L.CTRL,
      value = 1
    },
    {
      text = KEY:format(L.SHIFT),
      tooltipTitle = KEY:format(L.SHIFT),
      --tooltipText = KEYTIP:format(L.SHIFT),
	  key = L.SHIFT,
      value = 2
    },
    {
      text = L.MODIFIERKEY_NONE,
      tooltipTitle = L.MODIFIERKEY_NONE,
      tooltipText = L.MODIFIERKEY_NONE_TOOLTIP,
      value = 3
    };
  };

  --local scttext = L.INFORMONSKIP_SCT
  --if (not SCT) then  scttext = L.INFORMONSKIP_SCT_DISABLED:format(scttext);  end

  local panel, oldver = TjOptions.CreatePanel(THIS_TITLE, nil, {
  	title = THIS_TITLE.." v"..THIS_VERSION,
  	items = {
  		 { variable = "skipIsDefault", text = L.SKIPISDEFAULT, tooltip = L.SKIPISDEFAULT_TIP, tooltip2 = L.SKIPISDEFAULT_TIP2, OnChange = UpdateLabel },
  		 { variable = "greetingModifierKey", type = "dropdown", name = "Gossipmonger_GreetingModDropdown",
  		   text = L.MODIFIERKEY_LABEL_DONOT, menu = KeySelectDropDown, width = 85 },
  		 { variable = "showGossipMirrorOnSkip", text = L.SHOWBUBBLE, tooltip = L.SHOWBUBBLE_TIP },
  		 { variable = "delayShowingMirror", text = L.SHOWBUBBLE_DELAY,
  		   tooltip = L.SHOWBUBBLE_DELAY_TOOLTIP, xOffset = 14 },
  		 { variable = "showGossipMirrorOnQuestSelect", text = L.SHOWBUBBLE_QUESTSELECT,
  		   tooltip = L.SHOWBUBBLE_QUESTSELECT_TOOLTIP, tooltip2 = L.SHOWBUBBLE_QUESTSELECT_TOOLTIP2, xOffset = 0, OnChange = BreakMirror },
  		 { variable = "showGossipMirrorOnQuestJump", text = L.SHOWBUBBLE_QUESTJUMP,
  		   tooltip = L.SHOWBUBBLE_QUESTJUMP_TOOLTIP },

  		 { type = "label", text = L.FILTEROPT_STANDARD_LABEL, topBuffer = 10 },
  		 { variable = "Filter_Standard", text = L.FILTEROPT_STANDARD, tooltip = tooltip_StandardFilter,
  		   OnChange = UpdateStandardFilters },

  		 { type = "label", text = L.MISC_LABEL, topBuffer = 10 },
  		 { variable = "informOnSkip_log", type = "checkbox", text = L.INFORMONSKIP_CHAT },
  		 { variable = "alwaysInformWithName", type = "checkbox", text = L.INFORMONSKIP_NAMEONLY,
  		   tooltip = L.INFORMONSKIP_NAMEONLY_TOOLTIP, xOffset = 14 },

		   --[[
  		 { type = "label", text = L.INFORMONSKIP, xOffset = 0 },
  		 --{ variable = "informOnSkip_SCT", type = "checkboxsmall", text = scttext, xOffset = 14 },
  		 { variable = "informOnSkip_log", type = "checkboxsmall", text = L.INFORMONSKIP_CHAT, xOffset = 14 },
  		 { variable = "alwaysInformWithName", type = "checkboxsmall", text = L.INFORMONSKIP_NAMEONLY,
  		   tooltip = L.INFORMONSKIP_NAMEONLY_TOOLTIP },
		   --]]
  	},
  	variables = "Gossipmonger_Settings",
  	defaults = Gossipmonger.DefaultSettings,
  	OnShow = UpdateLabel
  });

  if (oldver and oldver ~= THIS_VERSION) then
    Gossipmonger_Settings.Version = THIS_VERSION
    for k,v in pairs(Gossipmonger.DefaultSettings) do
    -- Add options new to this version, set to default:
      if (Gossipmonger_Settings[k] == nil) then
        Gossipmonger_Settings[k] = v;
      end
    end
  end

  UpdateStandardFilters()

  -- Adjust position based on whether Quest NPC Model is shown:
  do
	local xoffset = UIPanelWindows.Gossipmonger_MirrorFrame.xoffset
	QuestNPCModel:HookScript("OnShow", function()
		SetUIPanelAttribute(Gossipmonger_MirrorFrame, "xoffset", (xoffset or 0) + QuestNPCModel:GetWidth())
		if (Gossipmonger_MirrorFrame:IsVisible()) then
			HideUIPanel(Gossipmonger_MirrorFrame)
			ShowUIPanel(Gossipmonger_MirrorFrame)
		end
	end)

	QuestNPCModel:HookScript("OnHide", function()
		SetUIPanelAttribute(Gossipmonger_MirrorFrame, "xoffset", xoffset)
		if (Gossipmonger_MirrorFrame:IsVisible()) then
			HideUIPanel(Gossipmonger_MirrorFrame)
			ShowUIPanel(Gossipmonger_MirrorFrame)
		end
	end)
  end

  Init = nil;
end

local function skipModifierKeyHeld()
  local key = Gossipmonger_Settings.greetingModifierKey
  if (key == 0) then
    return IsAltKeyDown()
  elseif (key == 1) then
    return IsControlKeyDown()
  elseif (key == 2) then
    return IsShiftKeyDown()
  end
  -- If key is set to None (3), it always returns nil.
end


local function buildOptionsTable(tab, isQuest, argsEach, ...)
  local list = tab.list and wipe(tab.list) or {}
  wipe(tab)
  local key2 = isQuest and "level" or "type"
  local count = 0
  for i = 1, select("#", ...), argsEach do
    count = count + 1
    local a, b, c = select(i, ...)
    tab["title"..count], list[#list+1] = a, a
    tab[key2..count], list[#list+1] = b, b
    if (isQuest) then
      c = not not c
      tab["isLowLevel"..count], list[#list+1] = c, c
    end
  end
  tab.list = list
end

local AvailableQuests, ActiveQuests, GossipOptions
local FILTERS = {}

local function findFilter(arg)
  if (type(arg) == "function") then
    for i,filter in ipairs(FILTERS) do
      if (filter.func == arg) then  return filter, i;  end
    end
  elseif (type(arg) == "string") then
    for i,filter in ipairs(FILTERS) do
      if (filter.name == arg) then  return filter, i;  end
    end
  end
end

local OptionClicked_type, OptionClicked_id, OptionClicked_expect_numOptions, OptionClicked_expect_title
local function OptionClickedFollowup(NPCName, text, numAvailable, available, numActive, active, numGossip, gossip)
  if (numAvailable + numActive + numGossip == OptionClicked_expect_numOptions) then
    local tab
    if (OptionClicked_type == 1) then
      tab = available
    elseif (OptionClicked_type == 2) then
      tab = active
    elseif (OptionClicked_type == 3) then
      tab = gossip
    else
      return;
    end
    if (tab["title"..OptionClicked_id] == OptionClicked_expect_title) then
      return OptionClicked_type, OptionClicked_id, nil, Gossipmonger_Settings.showGossipMirrorOnQuestJump
    end
  end
end


-- GLOBAL FUNCTIONS
-------------------------

function Gossipmonger.SetFilter(func, name, desc, priority, enabled)
-- Priority: Larger numbers go first. The default filter (which mimics standard WoW behavior) is set at priority 0.
-- Use nil for enabled to default to disabled if the filter didn't already exist or stay at whatever it is currently
-- set to if it does exist. Use nil for name, desc, and/or priority to use the current value(s). If the filter didn't
-- already exist, a default name (e.g. "Unnamed Filter"), a blank description, and a priority of 1 are used.
  assert(type(func) == "function", "Usage: Gossipmonger.SetFilter(func, name, desc, priority, enabled); func must be a function.")
  local tab, tabpos = findFilter(func)  -- See if filter is already used.

  local addtab
  if (tab) then
    name, desc = type(name) == "string" and name or tab.name, type(desc) == "string" and desc or tab.desc
    priority = type(priority) == "number" and priority or tab.priority
    if (enabled == nil) then  enabled = tab.enabled;  else  enabled = not not enabled;  end
    if (priority ~= tab.priority) then
      tremove(FILTERS, tabpos)
      addtab = true
    end
  else
    assert(priority == nil or type(priority) == "number", "Usage: Gossipmonger.SetFilter(func, name, desc, priority, enabled); priority must be nil or a number.")
    name = type(name) == "string" and name or L.FILTER_UNNAMED
    desc = type(desc) == "string" and desc or ""
    priority = priority or 1
    enabled = not not enabled
    tab, addtab = {}, true
  end

  tab.func, tab.name, tab.desc, tab.priority, tab.enabled = func, name, desc, priority, enabled

  if (addtab) then
    -- Find a suitable spot for a filter of this priority:
    for i,filter in ipairs(FILTERS) do
      if (filter.priority <= priority) then
        tinsert(FILTERS, i, tab)
        return;
      end
    end
    tinsert(FILTERS, tab)
  end
end

function Gossipmonger.GetFilterInfo(filter)
-- Accepts either function or name (string), but while a function can only be part of a single filter, it's possible
-- for a given name to be used by more than one.
  local tab = findFilter(filter)
  if (tab) then  return tab.func, tab.name, tab.desc, tab.priority, tab.enabled;  end
end

function Gossipmonger.EnableFilter(filter)
-- Accepts either function or name (string), but while a function can only be part of a single filter, it's possible
-- for a given name to be used by more than one.
  local tab = findFilter(filter)
  if (tab) then
    tab.enabled = true
    return true
  end
end

function Gossipmonger.DisableFilter(filter)
-- Accepts either function or name (string), but while a function can only be part of a single filter, it's possible
-- for a given name to be used by more than one.
  local tab = findFilter(filter)
  if (tab) then
    tab.enabled = false
    return true
  end
end

function Gossipmonger.OnLoad()
  tinsert(UISpecialFrames, "Gossipmonger_MirrorFrame");
  --UIPanelWindows.Gossipmonger_MirrorFrame = { area = "right", pushable = 8, whileDead = 1, yoffset = -10 };
  UIPanelWindows.Gossipmonger_MirrorFrame = { area = "right", pushable = 1, whileDead = 1, yoffset = -10 };
  Gossipmonger_MirrorFrame:SetScript("OnEvent", Init);
  Gossipmonger_MirrorFrame:RegisterEvent("VARIABLES_LOADED");
  Gossipmonger.OnLoad = nil
end

function Gossipmonger.OnEvent(self, event)
  if (event == "BANKFRAME_CLOSED" or event == "BATTLEFIELDS_CLOSED" or event == "MERCHANT_CLOSED" or
      event == "TAXIMAP_CLOSED" or event == "TRAINER_CLOSED" or event == "QUEST_FINISHED" or
      event == "GOSSIP_CLOSED" or event == "AUCTION_HOUSE_CLOSED") then
    if (event == "QUEST_FINISHED" and IgnoreQuestFinished and (IgnoreQuestFinished == true or GetTitleText() == IgnoreQuestFinished)) then
	  if (UnitExists("npc")) then  --if (QuestFrame:IsVisible() or GossipFrame:IsVisible()) then
	    IgnoreQuestFinished = nil
	    return;
	  end
    end
    BreakMirror()
  elseif (CurrentNPC == UnitName("npc")) then
    if (event == "QUEST_ACCEPTED" and not QuestFrameGreetingPanel:IsVisible()) then -- Looking at the last quest on this NPC after accepting another.
	  BreakMirror()
	  return;
	end
    ShowMirror()
	if (event == "QUEST_DETAIL") then  -- Looking at another quest, perhaps because we just picked one up, so watch for that so we can remove the bubble when this was the last quest:
	  Gossipmonger_MirrorFrame:RegisterEvent("QUEST_ACCEPTED")
	end
  else
    BreakMirror()
  end
end

function Gossipmonger.GoodbyeClick()
  IgnoreQuestFinished = nil
  if (MerchantFrame:IsShown()) then
    HideUIPanel(MerchantFrame)
  elseif (BankFrame:IsShown()) then
    HideUIPanel(BankFrame)
  --[[
  elseif (BattlefieldFrame:IsShown()) then
    HideUIPanel(BattlefieldFrame)
  elseif (ArenaFrame:IsShown()) then
    HideUIPanel(ArenaFrame)
  --]]
  elseif (PVEFrame:IsShown()) then -- Despite the name, this includes the PVP Queue.
    HideUIPanel(PVEFrame)
  elseif (TaxiFrame:IsShown()) then
    HideUIPanel(TaxiFrame)
  elseif (GossipFrame:IsShown()) then
    HideUIPanel(GossipFrame)
  elseif (QuestFrame:IsShown()) then
    HideUIPanel(QuestFrame)
  elseif (ClassTrainerFrame and ClassTrainerFrame:IsShown()) then
    HideUIPanel(ClassTrainerFrame)
  elseif (AuctionFrame and AuctionFrame:IsShown()) then
    HideUIPanel(AuctionFrame)
  end

  if (CurrentNPC_gossip == "binder") then
    local dialog = StaticPopup_FindVisible("CONFIRM_BINDER");
    if (dialog) then
      StaticPopup_OnClick(dialog, 2)   -- Call as if user pressed cancel button
    end
  end
  
  -- Failsafe:  (normally, closing the above will close out the mirror frame and break the connection)
  --BreakMirror() -- comment out until we find reproduceable steps showing this is really needed with the changes made to the rest of the code
end

function Gossipmonger.OptionClick(self)
  if (not SafeMirrorOptionsClick or self.selected) then  return;  end
  if (QuestFrame:IsShown()) then
    local optionType, typeid = self.type
    if (optionType == "Available") then
      typeid = 1
    elseif (optionType == "Active") then
      typeid = 2
    elseif (optionType == "Gossip") then
      typeid = 3
    elseif (self.QuestGreeting) then
      local icon = _G[self.baseButton.."QuestIcon"]
      if (icon:GetTexture() == "Interface\\GossipFrame\\ActiveQuestIcon") then
        typeid = 2
      else
        typeid = 1
      end
    else
      return;
    end

	IgnoreQuestFinished = true  -- Needed so using the hopping feature to select a vendor option will show the mirror once the vendor frame opens.
	-- (Although you can't make another hop from the vendor frame, showing it then can be good since the NPC's gossip text may have changed.)

    OptionClicked_type, OptionClicked_id = typeid, self:GetID()
    OptionClicked_expect_numOptions = Gossipmonger_MirrorFrame.numOptions
    OptionClicked_expect_title = self.title
    GossipNPC = UnitName("npc")
    FollowupFunction = OptionClickedFollowup
    FollowupFunction_filtername = "<QUESTJUMPER>"

    SafeMirrorOptionsClick = nil
    DeclineQuest();
    PlaySound("igQuestCancel");
    --HideUIPanel(Gossipmonger_MirrorFrame)  -- Hide the mirror (but don't "break" it)
  end
end

--[[ DEBUG: Force the "Gossip Bubble" (AKA Mirror Frame) to show:
function Gossipmonger.ShowBubble(delay, name, text, selected, gossipType)
  name = name or UnitName("npc")
  text = text and text:gsub("||n","|n") or GetGossipText()
  showGossipMirror(delay, name, text, selected, gossipType)
end
--]]

Gossipmonger.PrepGossipOptionDisplay = prepGossipOptionDisplay

Gossipmonger.ShowMirror = ShowMirror


-- HOOK
-------------------------

local orig_GossipFrame_OnEvent = GossipFrame_OnEvent

local function filtererror(filter, msg, ...)
  orig_GossipFrame_OnEvent(...)
  local name = filter and '"'..filter.name..'"' or '"'..FollowupFunction_filtername..'" followup'
  FollowupFunction = nil
  error("Invalid return from filter ("..name.."): "..msg, 2)
end

local function ShowGossipFrame(self)
  -- Show default gossip frame (see GossipFrame_OnEvent in FrameXML\GossipFrame.lua):
  if ( not GossipFrame:IsShown() ) then
    ShowUIPanel(self);
    if ( not self:IsShown() ) then
      CloseGossip();
      return;
    end
  end
  NPCFriendshipStatusBar_Update(self);
  GossipFrameUpdate();

  return true
end

GossipFrame_OnEvent = function(self, event, ...)
  if (event == "GOSSIP_SHOW") then
    --BreakMirror() -- needed here? It was found to be needed for QuestFrameGreetingPanel:OnShow hook.

    local name = UnitName("npc")
    local text = GetGossipText()
    local numAvailable, numActive, numGossip = GetNumGossipAvailableQuests(), GetNumGossipActiveQuests(), GetNumGossipOptions()

    local validFollowup, showMirrorIfNoFollowupResult
    if (GossipNPC == name or BlockNextFilter == name) then
    -- Dialog continued from auto-selected gossip option or player-canceled quest display:
      if (FollowupFunction and not BlockNextFilter) then
        validFollowup = FollowupFunction
        showMirrorIfNoFollowupResult = ShowMirrorOnGossip
      else
        if (not ShowGossipFrame(self)) then
          BreakMirror()
        elseif (BlockNextFilter) then
        -- This is the result of the player declining a quest or canceling the turn-in of a quest:
          if (numAvailable + numActive >= 1 and numAvailable + numActive + numGossip > 1) then
          -- If there are multiple options and at least one is a quest, prepare the mirror in case it's used by
          -- the Gossipmonger "quest hanger" feature:
            showGossipMirror(true, name, text, nil, nil, true, 1, numAvailable + numActive)
          else
            Gossipmonger_MirrorFrame.selected = nil
            Gossipmonger_MirrorFrame.numOptions = numAvailable + numActive + numGossip
            Gossipmonger_MirrorFrame.numQuests = numAvailable + numActive
            Gossipmonger_MirrorFrame.NPCName = name
          end
        elseif (ShowMirrorOnGossip) then
          ShowMirror()
          Gossipmonger_MirrorFrame:RegisterEvent("GOSSIP_CLOSED")
        end
        GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil
        return;
      end
    end
    GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil

    -- New dialog or using a followup function:
    local skipdefault, keyheld = Gossipmonger_Settings.skipIsDefault, skipModifierKeyHeld()
    if ( (validFollowup or (not keyheld and skipdefault) or (keyheld and not skipdefault)) and
         (numAvailable > 0 or numActive > 0 or numGossip > 0) ) then
      AvailableQuests, ActiveQuests, GossipOptions = AvailableQuests or {}, ActiveQuests or {}, GossipOptions or {}
      buildOptionsTable(AvailableQuests, true, 7, GetGossipAvailableQuests())
      buildOptionsTable(ActiveQuests, true, 6, GetGossipActiveQuests())
      buildOptionsTable(GossipOptions, false, 2, GetGossipOptions())
      local seltype, seloption, followup, override
      local thefilter

      if (validFollowup) then
        seltype, seloption, followup, override =
          validFollowup(name, text, numAvailable, AvailableQuests, numActive, ActiveQuests, numGossip, GossipOptions, true)
      else
        for i,filter in ipairs(FILTERS) do
          if (filter.enabled) then
            seltype, seloption, followup, override =
              filter.func(name, text, numAvailable, AvailableQuests, numActive, ActiveQuests, numGossip, GossipOptions)
            if (seltype and (seloption or seltype == 0)) then
              thefilter = filter
              break
            end
          end
        end
      end

      if (seltype and seloption and seltype ~= 0 and seloption ~= 0) then  -- Filter with results found:
        -- If seltype is 0, this section is skipped, causing the default gossip frame to be shown.
        if (type(seloption) ~= "number") then
          filtererror(thefilter, "Second return (option) must be a number.", self, event, ...)
          return;
        end
        local numOptions, SelectOption, gossipType
        if (seltype == 1) then
          numOptions, SelectOption = numAvailable, SelectGossipAvailableQuest
        elseif (seltype == 2) then
          numOptions, SelectOption = numActive, SelectGossipActiveQuest
        elseif (seltype == 3) then
          numOptions, SelectOption = numGossip, SelectGossipOption
          gossipType = GossipOptions["type"..seloption]
        else
          filtererror(thefilter, "First return (type) must be an integer between 0 and 3.", self, event, ...)
          return;
        end
        if (seloption < 1 or seloption > numOptions) then
          filtererror(thefilter, "Option of that type ("..seltype..") and index ("..seloption..") not found.", self, event, ...)
          return;
        end

        --local inform_SCT, inform_log = Gossipmonger_Settings.informOnSkip_SCT and SCT, Gossipmonger_Settings.informOnSkip_log
		local inform_log = Gossipmonger_Settings.informOnSkip_log
        --if (override ~= false and (inform_SCT or inform_log)) then
		if (override ~= false and not validFollowup and (inform_log)) then
          local named, inform = L.SKIPPINGTEXT_NAME:format(name)
          if (not gossipType or Gossipmonger_Settings.alwaysInformWithName or strlen(name) <= 20) then
            inform = named
          else
		    if (not typeReplacement) then
			  -- Creating lookup table
			  typeReplacement = {}
			  local arr = { strsplit("\n", L.NPCTYPE_REPLACE) }
			  for k,v in ipairs(arr) do
			    local t, r = strsplit("=", v, 2)
			    if (t and r and t ~= "" and r ~= "") then  typeReplacement[t] = r;  end
			  end
			end
            local titled = L.SKIPPINGTEXT_TYPE:format(typeReplacement[gossipType] and typeReplacement[gossipType] or gossipType)
            inform = strlen(named) <= strlen(titled) and named or titled
          end
          --if (inform_SCT) then
          --  SCTColors = SCTColors or { r = R, g = G, b = B }
          --  SCT:DisplayMessage(ChatIcon..inform, SCTColors)
          --end
          if (inform_log) then
            chatprint(ChatIcon.." "..CYAN..inform)
          end
        end

        -- If override is true, the mirror is forced to show. If it is false (not nil), it is forced not to.
        if (override == false) then
          BreakMirror()
          if (seltype ~= 3) then
            -- Prepare in case the "quest hanger" icon is clicked:
            showGossipMirror(true, name, text, index, gossipType, true, nil, numAvailable + numActive)
          end
        else
          local index = seloption
          if (seltype > 1) then  index = index + numAvailable;  end
          if (seltype > 2) then  index = index + numActive;  end
          if (override or Gossipmonger_Settings.showGossipMirrorOnSkip) then
            showGossipMirror(Gossipmonger_Settings.delayShowingMirror, name, text, index, gossipType, nil, (seltype ~= 3), numAvailable + numActive)
            ShowMirrorOnGossip = true
          elseif (seltype ~= 3) then
            -- Prepare in case the "quest hanger" icon is clicked:
            showGossipMirror(true, name, text, index, gossipType, true, 1, numAvailable + numActive)
          end
        end

        if (gossipType == "gossip") then
          GossipNPC = name
          if (type(followup) == "function") then
            FollowupFunction = followup
            FollowupFunction_filtername = thefilter.name
          end
        end
        --SelectOption(seloption)
		if (not validFollowup or ShowGossipFrame(self)) then
		  SelectOption(seloption)
		end
        return;

      elseif (showMirrorIfNoFollowupResult) then
        if (ShowGossipFrame(self)) then
          ShowMirror()
          Gossipmonger_MirrorFrame:RegisterEvent("GOSSIP_CLOSED")
        else
          BreakMirror()
        end
        return;
      end
    end

    local shown = ShowGossipFrame(self)

    if (Gossipmonger_MirrorFrame:IsShown()) then
    -- Fix the odd case where the mirror is sometimes left up after it was shown without a delay and the
    -- player ran away from the NPC before the selected option's result appears. (still relevant in latest version?)
      BreakMirror()
    end

    if (shown and numAvailable + numActive >= 1 and numAvailable + numActive + numGossip > 1) then
    -- If there are multiple options and at least one is a quest, prepare the mirror in case it's used by
    -- the Gossipmonger "quest hanger" feature:
      showGossipMirror(true, name, text, nil, nil, true, 1, numAvailable + numActive)
    else
      Gossipmonger_MirrorFrame.selected = nil
      Gossipmonger_MirrorFrame.numOptions = numAvailable + numActive + numGossip
      Gossipmonger_MirrorFrame.numQuests = numAvailable + numActive
      Gossipmonger_MirrorFrame.NPCName = name
    end

  else
    GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil
    orig_GossipFrame_OnEvent(self, event, ...)
  end
end

GossipFrame:SetScript("OnEvent", GossipFrame_OnEvent) -- Need to do this as of WoW 3.1 due to XML's function=<x> usage.

local orig_binderOnHide = StaticPopupDialogs["CONFIRM_BINDER"].OnHide;
StaticPopupDialogs["CONFIRM_BINDER"].OnHide = function(...)
  BreakMirror();
  if (orig_binderOnHide) then  orig_binderOnHide(...)  end
end

-- This function prevents Gossipmonger from running its filters when options appear due to the player hitting
-- Cancel buttons (e.g. "Cancel", "Decline", "Ignore", or "Unignore") to decline an offered quest or cancel turning in a quest.
local function QuestDeclined_DoNotFilter()
  BlockNextFilter = UnitName("npc")
end

--hooksecurefunc("QuestGoodbyeButton_OnClick", QuestDeclined_DoNotFilter)
--hooksecurefunc("QuestRewardCancelButton_OnClick", QuestDeclined_DoNotFilter)
--hooksecurefunc("QuestDetailDeclineButton_OnClick", QuestDeclined_DoNotFilter)

QuestFrameGoodbyeButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- "Cancel" when looking at in-progress quest
QuestFrameDeclineButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- "Decline" offered quest
--QuestFrameCancelButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- Button no longer there.
--QuestFrameDetailPanel.IgnoreButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- Button closes window so shouldn't have this hook.
--QuestFrameDetailPanel.UnignoreButton:HookScript("OnClick", QuestDeclined_DoNotFilter) -- Button closes window so shouldn't have this hook.
-- In fact, the other two buttons now close the window, so is this hook needed at all any more?

-- This function prevents the mirror from disappearing if it was visible when the player hits the "Continue"
-- button (for completing a quest).
local function QuestFrameShiftFromProgressToRewards()
  if (Gossipmonger_MirrorFrame:IsShown()) then
    IgnoreQuestFinished = GetTitleText()
  else
    IgnoreQuestFinished = nil
  end
end

--hooksecurefunc("QuestProgressCompleteButton_OnClick", QuestFrameShiftFromProgressToRewards)
QuestFrameCompleteButton:HookScript("OnClick", QuestFrameShiftFromProgressToRewards)


-- Handle the quest frame, which is used by a few NPCs which have no gossip (though many (most?) without gossip still use the gossip frame):
QuestFrameGreetingPanel:HookScript("OnShow", function()
  --print("QuestFrameGreetingPanel")
  BreakMirror()

  local name = UnitName("npc")
  local text = GetGreetingText()
  local numAvailable, numActive = GetNumAvailableQuests(), GetNumActiveQuests()

  local validFollowup = GossipNPC == name and FollowupFunction == OptionClickedFollowup and FollowupFunction
  GossipNPC, FollowupFunction, BlockNextFilter, ShowMirrorOnGossip = nil, nil, nil, nil

  if (validFollowup) then  -- Quest-jumping button clicked:
    AvailableQuests = AvailableQuests and wipe(AvailableQuests) or {}
    ActiveQuests = ActiveQuests and wipe (ActiveQuests) or {}
    GossipOptions = GossipOptions and wipe(GossipOptions) or {}

    local list = ActiveQuests.list and wipe(ActiveQuests.list) or {}
    local a, b, c
    for i = 1, numActive do
      a, b, c = GetActiveTitle(i), GetActiveLevel(i), IsActiveQuestTrivial(i)
      ActiveQuests["title"..i], list[#list+1] = a, a
      ActiveQuests["level"..i], list[#list+1] = b, b
      ActiveQuests["isLowLevel"..i], list[#list+1] = c, c
    end
    ActiveQuests.list = list

    list = AvailableQuests.list and wipe(AvailableQuests.list) or {}
    for i = 1, numAvailable do
      a, b, c = GetAvailableTitle(i), GetAvailableLevel(i), IsAvailableQuestTrivial(i)
      AvailableQuests["title"..i], list[#list+1] = a, a
      AvailableQuests["level"..i], list[#list+1] = b, b
      AvailableQuests["isLowLevel"..i], list[#list+1] = c, c
    end
    AvailableQuests.list = list

    local seltype, seloption, followup, override =
      validFollowup(name, text, numAvailable, AvailableQuests, numActive, ActiveQuests, 0, GossipOptions, true)

    if (seltype and seloption and seltype ~= 0 and seloption ~= 0) then
      -- If seltype is 0, this section is skipped, causing the default frame to be shown.
      if (type(seloption) ~= "number") then
        --filtererror(thefilter, "Second return (option) must be a number.", self, event, ...)
        local name = thefilter and '"'..thefilter.name..'"' or '"'..FollowupFunction_filtername..'" followup'
        error("Invalid return from filter ("..name.."): ".."Second return (option) must be a number.")
        return;
      end
      local numOptions, SelectOption, gossipType
      if (seltype == 1) then
        numOptions, SelectOption = numAvailable, SelectAvailableQuest
      elseif (seltype == 2) then
        numOptions, SelectOption = numActive, SelectActiveQuest
      elseif (seltype == 3) then
        numOptions, SelectOption = numGossip, SelectGossipOption
        gossipType = GossipOptions["type"..seloption]
      else
        --filtererror(thefilter, "First return (type) must be an integer between 0 and 3.", self, event, ...)
        local name = thefilter and '"'..thefilter.name..'"' or '"'..FollowupFunction_filtername..'" followup'
        error("Invalid return from filter ("..name.."): ".."First return (type) must be an integer between 0 and 3.")
        return;
      end
      if (seloption < 1 or seloption > numOptions) then
        --filtererror(thefilter, "Option of that type ("..seltype..") and index ("..seloption..") not found.", self, event, ...)
        local name = thefilter and '"'..thefilter.name..'"' or '"'..FollowupFunction_filtername..'" followup'
        error("Invalid return from filter ("..name.."): "..("Option of that type ("..seltype..") and index ("..seloption..") not found."))
        return;
      end

      --local inform_SCT, inform_log = Gossipmonger_Settings.informOnSkip_SCT and SCT, Gossipmonger_Settings.informOnSkip_log
	  local inform_log = Gossipmonger_Settings.informOnSkip_log
      --if (override ~= false and (inform_SCT or inform_log)) then
	  if (override ~= false and not validFollowup and (inform_log)) then
        local named, inform = L.SKIPPINGTEXT_NAME:format(name)
        if (not gossipType or Gossipmonger_Settings.alwaysInformWithName or strlen(name) <= 20) then
          inform = named
        else
          local titled = L.SKIPPINGTEXT_TYPE:format(gossipType)
          inform = strlen(named) <= strlen(titled) and named or titled
        end
        --if (inform_SCT) then
        --  SCTColors = SCTColors or { r = R, g = G, b = B }
        --  SCT:DisplayMessage(ChatIcon..inform, SCTColors)
        --end
        if (inform_log) then
          chatprint(ChatIcon.." "..CYAN..inform)
        end
      end

      -- If override is true, the mirror is forced to show. If it is false (not nil), it is forced not to.
      if (override == false) then
        BreakMirror()
        if (seltype ~= 3) then
          -- Prepare in case the "quest hanger" icon is clicked:
          showGossipMirror(true, name, text, index, gossipType, true, nil, numAvailable + numActive, true)
        end
      else
        local index = seloption
        if (seltype == 1) then  index = index + numAvailable;  end
          -- Note that this is different from the normal method because active quests are displayed before available
          -- quests here (the reverse of normal).
        if (override or Gossipmonger_Settings.showGossipMirrorOnSkip) then
          showGossipMirror(Gossipmonger_Settings.delayShowingMirror, name, text, index, gossipType, nil, (seltype ~= 3), numAvailable + numActive, true)
          ShowMirrorOnGossip = true
        elseif (seltype ~= 3) then
          -- Prepare in case the "quest hanger" icon is clicked:
          showGossipMirror(true, name, text, index, gossipType, true, 1, numAvailable + numActive, true)
        end
      end

      if (gossipType == "gossip") then
        GossipNPC = name
        if (type(followup) == "function") then
          FollowupFunction = followup
          FollowupFunction_filtername = thefilter.name
        end
      end
      SelectOption(seloption)
      --[[ NEED TO TEST whether the line above alone is enough. It used to be enough for the GossipFrame version, but it stopped working so the below was
           added to that function. But it's likely that if the line above isn't enough, the code below won't be what we need, either, since it's for the
           gossip frame, not QuestFrameGreetingPanel. Some adaptation of it could work, though (e.g. forcing QuestFrameGreetingPanel to show/update?).
	  if (not validFollowup or ShowGossipFrame(self)) then
	    SelectOption(seloption)
	  end
	  --]]
      return;
    end
  end

  if (numAvailable + numActive >= 1 and numAvailable + numActive > 1) then
  -- If there are multiple quests, prepare the mirror in case it's used by the Gossipmonger "quest hanger" feature:
    showGossipMirror(true, name, text, nil, nil, true, 1, numAvailable + numActive, true)
	Gossipmonger_MirrorFrame.NPCName = name
  else
    Gossipmonger_MirrorFrame.selected = nil
    Gossipmonger_MirrorFrame.numOptions = numAvailable + numActive
    Gossipmonger_MirrorFrame.numQuests = numAvailable + numActive
    Gossipmonger_MirrorFrame.NPCName = name
  end
end)


-- TEST LOCATION: Refuge Pointe. Skuerto (NPC) and Wanted Board (object).


